import logging
import sys

from .winrm import WinRMProtocol
from .ssh import SSHProtocol

try:
    from .ldap import LDAPProtocol
except ImportError:
    logging.warning('For the LDAP protocol to work correctly, install the dependency:')
    if 'win' in sys.platform.lower():
        logging.warning('\tpip install https://github.com/cgohlke/python-ldap-build/releases/download/v3.4.4/python_ldap-3.4.4-cp38-cp38-win_amd64.whl')
    else:
        logging.warning('\tpip install python-ldap==3.4.4')

INTERFACES = {
    'ssh': SSHProtocol,
    # 'winexe': WinexeProtocol,
    # 'wmi': WMIProtocol,
    'winrm': WinRMProtocol,
    'ldap': LDAPProtocol,
}


from abc import ABC, abstractmethod

from ..model import Response, Query, Protocol


class BaseProtocol(ABC):
    @abstractmethod
    def connect(self, protocol: Protocol) -> None:
        raise NotImplementedError('Метод connect не реализован')

    @abstractmethod
    def execute(self, query: Query) -> Response:
        raise NotImplementedError('Метод execute не реализован')

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError('Метод close не реализован')



import io
import socket
import time
from typing import Optional

import paramiko
from paramiko.rsakey import RSAKey
from paramiko.ssh_exception import AuthenticationException, SSHException

from ..model import Protocol, Query, Response
from .base import BaseProtocol
from ..exceptions import ConnectionFailed, ExecutionError, CredentialsError


class SSHProtocol(BaseProtocol):
    name = 'ssh'
    """
    A class used to represent a ssh connection to remote host

    :param host: hostname or ip address of remote server
    :param login: user login for remote connection
    :param password: user password for remote connection
    :param port: ssh port. default 22 If not specified
    :param SSH_CONNECT_TIMEOUT: an optional timeout (in seconds)
        for the TCP connect
    :param EXEC_COMMAND_TIMEOUT: an optional timeout (in seconds)
        for execute command
    """

    SSH_CONNECT_TIMEOUT: int = 60
    EXEC_COMMAND_TIMEOUT: int = 60

    def __init__(self, protocol: Optional[Protocol] = None):
        if not self._connection and protocol:
            self.connect(protocol)

    def connect(self, protocol: Protocol):
        """Open SSH connection to remote host.

        :raises: CreateSSHConnectionError: if failed authentication
            or connection or establishing an SSH session
        """
        host = protocol.host
        login = protocol.username
        password = protocol.password
        port = protocol.port
        ssh_key = protocol.pem_file
        try:
            if password:
                self._connection = paramiko.SSHClient()
                self._connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self._connection.connect(
                    hostname=host, username=login,
                    password=password, port=port,
                    look_for_keys=False, allow_agent=False,
                    timeout=SSHProtocol.SSH_CONNECT_TIMEOUT
                )
            elif ssh_key:
                keyfile = io.StringIO(ssh_key)
                self._connection = paramiko.SSHClient()
                self._connection.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                self._connection.connect(
                    host, port, login,
                    pkey=RSAKey.from_private_key(keyfile),
                    timeout=SSHProtocol.SSH_CONNECT_TIMEOUT,
                    look_for_keys=False,
                    allow_agent=False
                )
        except AuthenticationException as error:
            raise CredentialsError(str(error))
        except (SSHException, socket.timeout) as exc:
            raise ConnectionFailed(str(exc))
        except Exception as error:
            raise ConnectionFailed(f'SSH connection failed: {error}')

    def close(self):
        """Close SSH connection."""
        if self._connection is not None:
            self._connection.close()
            self._connection = None

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, exc_type: Exception, *args):
        self.close()
        if exc_type is not None:
            return False
        return True

    def execute(self, query: Query) -> Response:
        command = query.command
        sudo = query.sudo
        timeout = query.timeout
        if self._connection is None:
            self.connect()
        stdin, stdout, err = self._connection.exec_command(
            command=command,
            timeout=timeout)
        if sudo:
            stdin.write(sudo + "\n")
            stdin.flush()
            time.sleep(1)
        error = err.read().decode('utf-8')
        if not stdout and error:
            raise ExecutionError(error + stdout.read().decode('utf-8') + f'\nCommand: {command}')
        else:
            return Response(result=stdout.read().decode('utf-8'), status=True, text='Command executed successfully')



import json
from typing import List, Union


class SOARException(Exception):
    """
    Стандартный класс для исключений скриптов.

    Приватные поля:
        ``_return_code``(int): код возврата
        ``_TEMPLATE``(str): шаблон сообщения

    Публичные поля и свойства:
        ``message``(Union[List[str], str, dict]): информация об исключении
        ``return_code``(int): код возврата
    """
    _return_code: int = 1
    _TEMPLATE: str = 'Error: {args}.'

    message: str = None

    @property
    def return_code(self):
        return self._return_code

    def __init__(self, message: Union[List[str], str, dict, Exception]):
        if message is None:
            self.message = ' '
        if isinstance(message, Exception):
            self.message = str(message)
        elif isinstance(message, list):
            self.message = ', '.join(message)
        elif isinstance(message, dict):
            self.message = json.dumps(message)
        elif isinstance(message, str):
            self.message = message
        else:
            raise TypeError('Message must be str, list, dict or Exception')

    def __str__(self):
        return self._TEMPLATE.format(args=self.message)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
import logging
import sys

from setuptools import setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()


if 'win' in sys.platform.lower():
    logging.info('For the LDAP protocol to work correctly, install the dependency:')
    logging.info('\tpip install https://github.com/cgohlke/python-ldap-build/releases/download/v3.4.4/python_ldap-3.4.4-cp38-cp38-win_amd64.whl')

setup(
    name='ussl',
    author='ussc soc dev team',
    author_email='iushangaraev@ussc.ru, pbikkuzhina@ussc.ru',
    description='Пакет разработчиков USSC-SOC для упрощения взаимодействия с АРМ, серверами и сетевыми устройствами',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='1.1.1',
    python_requires='>=3.8.0',
    packages=[
        'ussl',
        'ussl.model',
        'ussl.postprocessing',
        'ussl.protocol',
        'ussl.transport',
        'ussl.exceptions',
        'ussl.utils',
    ],
    install_requires=[
        'pywinrm==0.4.1',
        'paramiko==2.7.2',
        'marshmallow==3.20.2',
        'python-ldap==3.4.4; platform_system != "Windows"',
        # 'python-ldap-build @ https://github.com/cgohlke/python-ldap-build/releases/download/v3.4.4/python_ldap-3.4
        # .4-cp38-cp38-win_amd64.whl', # for Windows only
    ],
    classifiers=[
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.8'
    ]
)

# To update version:
# python setup.py sdist
# twine upload dist/*



import re
import ssl
from typing import Any, Union

import requests
from requests.adapters import HTTPAdapter
from requests.exceptions import ReadTimeout
from urllib3 import ProxyManager
from urllib3.util.ssl_ import create_urllib3_context
from winrm import Session
from winrm.exceptions import (
    AuthenticationError,
    WinRMError,
    WinRMOperationTimeoutError,
    WinRMTransportError,
)

from .base import BaseProtocol
from ..model import Response, Protocol, Query
from ..exceptions import ConnectionFailed, ExecutionError, CredentialsError, PermissionsError


class TLSv1Adapter(HTTPAdapter):
    """
    Включаем поддержку TLSv1.0 для старых версий Windows.
    """

    def init_poolmanager(self, *args: Any, **kwargs: Any) -> None:
        context = create_urllib3_context()
        context.minimum_version = ssl.TLSVersion.TLSv1
        context.options &= (
                ~ssl.OP_NO_TLSv1_3  # pylint: disable=no-member
                & ~ssl.OP_NO_TLSv1_2
                & ~ssl.OP_NO_TLSv1_1
                & ~ssl.OP_NO_TLSv1
        )
        kwargs['ssl_context'] = context
        return super().init_poolmanager(*args, **kwargs)

    def proxy_manager_for(self, *args: Any, **kwargs: Any) -> ProxyManager:
        context = create_urllib3_context()
        context.minimum_version = ssl.TLSVersion.TLSv1
        context.options &= (
                ~ssl.OP_NO_TLSv1_3  # pylint: disable=no-member
                & ~ssl.OP_NO_TLSv1_2
                & ~ssl.OP_NO_TLSv1_1
                & ~ssl.OP_NO_TLSv1
        )
        kwargs['ssl_context'] = context
        return super().proxy_manager_for(*args, **kwargs)


class WinRMProtocol(BaseProtocol):
    name = 'winrm'
    WINRM_PORT = 5985
    WINRM_SSL_PORT = 5986
    VALID_TRANSPORT = {'plaintext', 'ntlm'}

    def __init__(self) -> None:
        self._session: Union[Session, None] = None

    def connect(
            self,
            protocol: Protocol,
    ) -> None:
        host = protocol.host
        username = protocol.username
        domain = protocol.domain
        password = protocol.password
        scheme = protocol.default('scheme', 'https')
        if scheme == 'https':
            port = protocol.default('port', self.WINRM_SSL_PORT)
        else:
            port = protocol.default('port', self.WINRM_SSL_PORT)
        path = protocol.default('path', 'wsman')
        # Отправка запроса в кодировке 437 даёт ответ на анлийском языке
        # однако при преобразовании байтов в строку и использованием той же
        # кодировки русские символы отображаются некорректно. Для того чтобы
        # обойти эту проблему, для декодирования используется кодировка 866
        # encoding = protocol.default('encoding', 437)
        self._decoding = protocol.default('decoding', 866)
        window_width = protocol.default('window_width', 300)
        transport = protocol.default('transport', 'ntlm')

        if transport not in self.VALID_TRANSPORT:
            raise ConnectionFailed(f'{transport} protocol is not supported')

        # для BasicAuth явно преобразуем 'username' и 'password' из utf8 в cp1251, т.к. requests.auth.HTTPBasicAuth
        # по умолчанию преобразует utf8 строки в latin1, при этом bytes оставляет 'как есть'.
        # https://github.com/requests/requests/pull/3673
        _username: Union[str, bytes] = f'{domain}\\{username}'
        _password: Union[str, bytes] = password
        if transport == 'plaintext':
            _username = f'{domain}\\{username}'.encode('cp1251')
            _password = password.encode('cp1251')
        else:
            _username = f'{domain}\\{username}'
            _password = password

        endpoint = f'{scheme}://{host}:{port}/{path}'
        try:
            self._session = Session(f'https://{host}:5986',
                                    auth=(_username, _password),
                                    server_cert_validation='ignore',
                                    transport=transport)

            self._session.protocol.transport.session.mount(endpoint, TLSv1Adapter())
            self.execute(Query(command=f'mode con:cols={window_width}', timeout=10))
        except requests.exceptions.ConnectionError as e:
            raise ConnectionFailed(str(e))

    def close(self) -> None:
        self._session = None

    def execute(
            self,
            query: Query
    ) -> Response:
        try:
            if query.shell_type == 'ps':
                response = self._session.run_ps(query.command)
            else:
                response = self._session.run_cmd(query.command)
            std_out, std_err, status_code = response.std_out, response.std_err, response.status_code
            if status_code != 0 and status_code == 2147942405:
                raise PermissionsError(std_out.decode(encoding=str(self._decoding), errors='ignore'))
            if status_code != 0:
                std_err = std_err.decode(encoding=str(self._decoding), errors='ignore').strip()
                raise ExecutionError(std_err)

            std_out = std_out.decode(encoding=str(self._decoding)).strip()
            return Response(
                result=std_out,
                text=f'Command {query.command} executed successfully',
                status=True)

        except WinRMTransportError as exc:
            raise ConnectionFailed(exc.__str__())

        except AuthenticationError as exc:
            raise CredentialsError(exc.__str__())

        except (WinRMOperationTimeoutError, ReadTimeout):
            raise ConnectionFailed('Connection timeout')

        except WinRMError as exc:
            raise ExecutionError(exc.__str__())
